use std::rc::Rc;

use im::hashmap::*;
use wasm_bindgen::prelude::*;
use web_sys::{Element, Node};
use yew::prelude::*;

#[wasm_bindgen]
extern "C" {
    #[derive(PartialEq)]
    pub type Terminal;
    #[wasm_bindgen(constructor)]
    fn new() -> Terminal;
    #[wasm_bindgen(method)]
    fn open(this: &Terminal, element: &Element);
    #[wasm_bindgen(method)]
    fn write(this: &Terminal, data: &str);
    // Todo: Can we do this with interfaces somehow?
    #[wasm_bindgen(method)]
    fn loadAddon(this: &Terminal, addon: &FitAddon);
    #[derive(PartialEq)]
    pub type FitAddon;
}

#[wasm_bindgen(js_namespace=FitAddon)]
extern "C" {
    #[wasm_bindgen(constructor)]
    fn new() -> FitAddon;
    #[wasm_bindgen(method)]
    fn fit(this: &FitAddon);
}

#[derive(Eq, Ord, Hash, PartialEq, PartialOrd, Clone)]
pub struct TermFrame(pub u64);

#[derive(Properties, PartialEq)]
pub struct TermFrameData {
    pub id: TermFrame,
    pub term: Terminal,
    pub fit: FitAddon,
    pub node: Node,
}

pub type RegisteredTermFrames = HashMap<TermFrame, Rc<TermFrameData>>;
#[derive(Properties, PartialEq, Clone)]
pub struct RegisteredTermFrameLens {
    pub get: Rc<RegisteredTermFrames>,
    pub set: Callback<Rc<RegisteredTermFrames>, ()>,
}

fn get_or_make_term_frame(
    frame: &TermFrame,
    frames: &RegisteredTermFrameLens,
) -> Rc<TermFrameData> {
    if let Some(tfd) = frames.get.get(frame) {
        return tfd.clone();
    }

    let mut new_frames: RegisteredTermFrames = (*frames.get).clone();
    let term = Terminal::new();
    let fit = FitAddon::new();
    let element = web_sys::window()
        .and_then(|w| w.document())
        .and_then(|d| d.create_element("div").ok())
        .expect("Can create element for term");
    element.set_class_name("hterminal");
    term.open(&element);
    term.loadAddon(&fit);
    for i in 0..100 {
        term.write(&format!("{} Hello world\r\n", i));
    }
    let new_data: Rc<TermFrameData> = TermFrameData {
        id: frame.clone(),
        term,
        fit,
        node: element.into(),
    }
    .into();
    new_frames.insert(frame.clone(), new_data.clone());
    frames.set.emit(new_frames.into());
    new_data
}

#[derive(Properties, PartialEq)]
pub struct TermViewProps {
    pub terminal: TermFrame,
    pub frames: RegisteredTermFrameLens,
}

#[function_component(TermView)]
pub fn term_view(props: &TermViewProps) -> Html {
    let term = get_or_make_term_frame(&props.terminal, &props.frames);
    term.fit.fit();
    html! {
        {Html::VRef(term.node.clone())}
    }
}
